from django.db import models

from edge.importer import GFFImporter
from edge.models.chunk import (
    Annotation,
    Chunk_Feature,
    Edge,
    Fragment_Chunk_Location,
)
from edge.models.fragment import Fragment
from edge.models.genome_updater import Genome_Updater


class Genome(Genome_Updater, models.Model):
    class Meta:
        app_label = "edge"

    name = models.TextField()
    parent = models.ForeignKey(
        "self", null=True, on_delete=models.PROTECT, related_name="children"
    )
    notes = models.TextField(null=True, blank=True)
    fragments = models.ManyToManyField(Fragment, through="Genome_Fragment")
    created_on = models.DateTimeField("Created", auto_now_add=True, null=True)
    active = models.BooleanField(default=True)
    blastdb = models.TextField(null=True, blank=True)

    def __str__(self):
        return self.name

    @staticmethod
    def create(name, notes=None):
        new_genome = Genome(name=name, notes=notes, parent=None)
        new_genome.save()
        return new_genome

    @staticmethod
    def import_gff(name, gff_fasta_fn, dirn='.'):
        genome = Genome.create(name)
        GFFImporter(genome, gff_fasta_fn).do_import(dirn=dirn)
        return genome

    def update(self, name=None, notes=None):
        name = self.name if name is None else name
        new_genome = Genome(name=name, notes=notes, parent=self)
        new_genome.save()
        for f in self.fragments.all():
            Genome_Fragment(genome=new_genome, fragment=f, inherited=True).save()
        return new_genome

    @property
    def has_location_index(self):
        for f in self.fragments.all():
            if not f.has_location_index:
                return False
        return True

    def indexed_genome(self):
        for f in sorted(self.fragments.all(), key=lambda f: f.id):
            f.indexed_fragment()
        # casting to Indexed_Genome
        return Indexed_Genome.objects.get(pk=self.pk)

    def root(self):
        if self.parent is None:
            return self
        return self.parent.root()

    def lock(self):
        """
        Do a select for update, which, when executed inside a transaction, places
        a lock on the genome.
        """

        root = self.root()

        genomes = Genome.objects.select_for_update().filter(pk=root.id)
        # Lock only happens when queryset is evaluated, therefore need to do at least genomes[0]
        genome = genomes[0]
        print(f"Lock genome {genome.id}")
        return self


class Indexed_Genome(Genome):
    """
    An Indexed_Genome is a Genome where each fragment has been indexed.
    """

    class Meta:
        app_label = "edge"
        proxy = True

    def __annotations_from_chunk_features(self, chunk_features):
        cf_fcl = []
        for cf in chunk_features:
            # get fragment chunk location
            for fcl in Fragment_Chunk_Location.objects.filter(
                chunk=cf.chunk, fragment__genome=self
            ):
                cf_fcl.append((cf, fcl))
        annotations = Annotation.from_chunk_feature_and_location_array(cf_fcl)

        by_f = {}
        for annotation in annotations:
            if annotation.fragment.id not in by_f:
                by_f[annotation.fragment.id] = []
            by_f[annotation.fragment.id].append(annotation)
        return by_f

    def find_annotation_by_feature(self, feature):
        q = Chunk_Feature.objects.filter(
            chunk__fragment_chunk_location__fragment__genome=self, feature=feature
        )
        return self.__annotations_from_chunk_features(list(q))

    def find_annotation_by_name(self, name):
        q = Chunk_Feature.objects.filter(
            chunk__fragment_chunk_location__fragment__genome=self, feature__name=name
        )
        return self.__annotations_from_chunk_features(list(q))

    def find_annotation_by_qualifier(self, name, fields=None):
        fields = None if fields is None else [f.lower() for f in fields]

        q = Chunk_Feature.objects.filter(
            chunk__fragment_chunk_location__fragment__genome=self,
            feature___qualifiers__icontains=name,
        )
        chunk_features = []
        for cf in q:
            qualifiers = cf.feature.qualifiers
            for k, v in qualifiers.items():
                if fields is None or k.lower() in fields:
                    if type(v) in (bytes, str):
                        v = v.split(",")
                    v = [s.lower() for s in v]
                    if name.lower() in v:
                        chunk_features.append(cf)

        return self.__annotations_from_chunk_features(chunk_features)

    def changes(self):
        if self.parent is None:
            return []

        edges = Edge.objects.filter(
            fragment__genome_fragment__genome=self,
            fragment__genome_fragment__inherited=False,
        )

        fcs = []
        added = []
        for edge in edges:
            if edge.from_chunk:
                if edge.from_chunk.id not in added:
                    fcs.append(
                        edge.fragment.indexed_fragment().fragment_chunk(edge.from_chunk)
                    )
                    added.append(edge.from_chunk.id)
            if edge.to_chunk:
                if edge.to_chunk.id not in added:
                    fcs.append(
                        edge.fragment.indexed_fragment().fragment_chunk(edge.to_chunk)
                    )
                    added.append(edge.to_chunk.id)
        return fcs

    def changed_locations_by_fragment(self):
        changes = {}

        for c in self.changes():
            if c.fragment.id not in changes:
                changes[c.fragment.id] = []
            v = changes[c.fragment.id]
            if len(v) == 0 or v[-1][1] + 1 != c.location[0]:
                v.append([c.location[0], c.location[1]])
            else:
                v[-1][1] = c.location[1]

        changes = {Fragment.objects.get(pk=f): v for f, v in changes.items()}
        return changes

    def get_coordinate_diff_from_parent_genome(self, p_genome_id):
        # Get fragments contained by child and not by parent
        p_fragment_ids = Fragment.objects.filter(genome=p_genome_id).values_list('id', flat=True)
        child_only_fragments = self.fragments.filter().exclude(genome=p_genome_id)

        # Relate child fragments to parent fragments
        child_to_parent_fs = {}
        for f in child_only_fragments:
            curr_f = f
            while curr_f is not None and curr_f.id not in p_fragment_ids:
                curr_f = curr_f.parent
            if curr_f is not None:
                child_to_parent_fs[f.indexed_fragment()] = curr_f.indexed_fragment()
        if len(child_to_parent_fs) == 0:
            raise Exception("Genome input is not a parent by lineage")

        # Get start to end chunks by fragment of diff regions between parent and child
        fragment_to_starts_and_ends = {}
        for child_f, parent_f in child_to_parent_fs.items():
            parent_chunks, child_chunks = list(parent_f.chunks()), list(child_f.chunks())
            starts_and_ends, current_start = {}, None
            for child_i, c in enumerate(child_chunks):
                if c in parent_chunks:
                    parent_i = parent_chunks.index(c)
                    prev_parent_chunk = parent_chunks[parent_i - 1] \
                        if parent_i > 0 else None
                    next_parent_chunk = parent_chunks[parent_i + 1] \
                        if parent_i < (len(parent_chunks) - 1) else None
                    prev_child_chunk = child_chunks[child_i - 1] \
                        if child_i > 0 else None
                    next_child_chunk = child_chunks[child_i + 1] \
                        if child_i < (len(child_chunks) - 1) else None

                    if prev_parent_chunk != prev_child_chunk:
                        if current_start is not None:
                            starts_and_ends[current_start] = c.id
                            current_start = None
                        if len(starts_and_ends) == 0:
                            starts_and_ends[None] = c.id

                    if next_parent_chunk != next_child_chunk:
                        current_start = c.id

            if current_start is not None:
                starts_and_ends[current_start] = None
            fragment_to_starts_and_ends[child_f] = starts_and_ends

        # Store dictionaries of changes with ID, start, and end for parent and child fragments
        regions = []
        for child_f, starts_and_ends in fragment_to_starts_and_ends.items():
            parent_f = child_to_parent_fs[child_f]
            for start_id, end_id in starts_and_ends.items():
                regions.append({
                    "parent_fragment_name": parent_f.name,

                    "parent_fragment_id": parent_f.id,

                    "parent_starts_at": 1 if start_id is None
                    else parent_f.fragment_chunk(start_id).base_last + 1,

                    "parent_ends_before": parent_f.length + 1 if end_id is None
                    else parent_f.fragment_chunk(end_id).base_first,

                    "child_fragment_name": child_f.name,

                    "child_fragment_id": child_f.id,

                    "child_starts_at": 1 if start_id is None
                    else child_f.fragment_chunk(start_id).base_last + 1,

                    "child_ends_before": child_f.length + 1 if end_id is None
                    else child_f.fragment_chunk(end_id).base_first,
                })

        return regions


class Genome_Fragment(models.Model):
    class Meta:
        app_label = "edge"

    genome = models.ForeignKey(Genome, on_delete=models.CASCADE)
    fragment = models.ForeignKey(Fragment, on_delete=models.CASCADE)
    inherited = models.BooleanField()



import uuid
import json

from django.test import TestCase
from edge.ssr.crelox import Sites
from edge.models import Genome, Fragment, Genome_Fragment


class SSRViewTest(TestCase):

    def test_check_api_without_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            Sites.loxP + "t" * 100 + Sites.loxP + "c" * 1000
        )
        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        res = self.client.post(
            "/edge/genomes/%s/ssr/" % parent_genome.id,
            data=json.dumps(
                dict(donor=None, is_donor_circular=None, reaction="crelox", create=False)
            ),
            content_type="application/json",
        )
        self.assertEquals(res.status_code, 200)
        r = json.loads(res.content)

        self.assertEquals(len(r), 1)
        self.assertEquals(r[0]["recombination"]["type"], "Excision")
        self.assertEquals(len(r[0]["genomic_locations"]), 2)
        self.assertEquals(r[0]["genomic_locations"][0]["fragment_id"], f.id)
        self.assertEquals(r[0]["genomic_locations"][0]["start"], 1)
        self.assertEquals(r[0]["genomic_locations"][1]["fragment_id"], f.id)
        self.assertEquals(r[0]["genomic_locations"][1]["start"], len(Sites.loxP) + 100 + 1)

    def test_check_api_with_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 100 + Sites.lox71 + "c" * 1000
        )
        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        donor = "a" * 10 + Sites.lox66 + "g" * 100

        res = self.client.post(
            "/edge/genomes/%s/ssr/" % parent_genome.id,
            data=json.dumps(
                dict(donor=donor, is_donor_circular=True, reaction="crelox", create=False)
            ),
            content_type="application/json",
        )
        self.assertEquals(res.status_code, 200)
        r = json.loads(res.content)

        self.assertEquals(len(r), 1)
        self.assertEquals(r[0]["recombination"]["type"], "Integration")
        self.assertEquals(len(r[0]["genomic_locations"]), 1)
        self.assertEquals(r[0]["genomic_locations"][0]["fragment_id"], f.id)
        self.assertEquals(r[0]["genomic_locations"][0]["start"], 101)

    def test_create_genome_api_without_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            Sites.loxP + "t" * 100 + Sites.loxP + "c" * 1000
        )
        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        res = self.client.post(
            "/edge/genomes/%s/ssr/" % parent_genome.id,
            data=json.dumps(
                dict(donor=None, is_donor_circular=None, reaction="crelox",
                     genome_name="foo foo bar bar", create=True)
            ),
            content_type="application/json",
        )
        self.assertEquals(res.status_code, 201)
        r = json.loads(res.content)

        child = Genome.objects.get(pk=r["id"])
        self.assertEquals(child.name, "foo foo bar bar")
        modified_sequence = child.fragments.all()[0].indexed_fragment().sequence
        self.assertEquals("t" * 100 not in modified_sequence, True)

    def test_create_genome_api_with_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 100 + Sites.lox71 + "c" * 1000
        )
        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        donor = "a" * 10 + Sites.lox66 + "g" * 100

        res = self.client.post(
            "/edge/genomes/%s/ssr/" % parent_genome.id,
            data=json.dumps(
                dict(donor=donor, is_donor_circular=True, reaction="crelox",
                     genome_name="foo foo bar bar", create=True)
            ),
            content_type="application/json",
        )
        self.assertEquals(res.status_code, 201)
        r = json.loads(res.content)

        child = Genome.objects.get(pk=r["id"])
        self.assertEquals(child.name, "foo foo bar bar")
        modified_sequence = child.fragments.all()[0].indexed_fragment().sequence
        self.assertEquals("g" * 100 in modified_sequence, True)


class SSRTester(TestCase):

    def setUp(self):
        self.genome = None
        self.result = None

    def with_genome(self, sequence):
        genome = Genome(name=str(uuid.uuid4()))
        genome.save()
        f = Fragment.create_with_sequence(str(uuid.uuid4()), sequence)
        Genome_Fragment(genome=genome, fragment=f, inherited=False).save()
        self.genome = genome
        return self

    def when_trigger(self, reaction_name, donor=None, is_donor_circular=True):
        res = self.client.post(
            "/edge/genomes/%s/ssr/" % self.genome.id,
            data=json.dumps(
                dict(
                    donor=donor, is_donor_circular=is_donor_circular,
                    reaction=reaction_name, create=True
                )
            ),
            content_type="application/json",
        )
        self.result = res
        return self

    def when_trigger_crelox(self, donor=None, is_donor_circular=None):
        return self.when_trigger("crelox", donor=donor, is_donor_circular=is_donor_circular)

    def when_trigger_crelox_with_donor(self, donor, is_donor_circular=True):
        return self.when_trigger_crelox(donor=donor, is_donor_circular=is_donor_circular)

    def when_trigger_flp(self, donor=None, is_donor_circular=None):
        return self.when_trigger("flp", donor=donor, is_donor_circular=is_donor_circular)

    def when_trigger_flp_with_donor(self, donor, is_donor_circular=True):
        return self.when_trigger_flp(donor=donor, is_donor_circular=is_donor_circular)

    def modified_sequence_is(self, s):
        self.assertEquals(self.result.status_code, 201)

        result = json.loads(self.result.content)
        child = Genome.objects.get(pk=result["id"])
        modified_sequence = child.fragments.all()[0].indexed_fragment().sequence.lower()
        s = s.lower()

        if modified_sequence != s:
            print(modified_sequence)
            print(s)
        self.assertEquals(modified_sequence, s)
        return self

    def is_not_allowed(self):
        self.assertEquals(self.result.status_code, 400)



from edge.ssr import rc, lower_no_whitespace, Reaction, Integration, Excision, Inversion, RMCE


class Sites(object):
    loxP = lower_no_whitespace("ATAACTTCGTATA GCATACAT TATACGAAGTTAT")
    lox66 = lower_no_whitespace("ATAACTTCGTATA GCATACAT TATACGAACGGTA")
    lox71 = lower_no_whitespace("TACCGTTCGTATA GCATACAT TATACGAAGTTAT")
    lox72 = lower_no_whitespace("TACCGTTCGTATA GCATACAT TATACGAACGGTA")

    lox5171 = lower_no_whitespace("ATAACTTCGTATA GtAcACAT TATACGAAGTTAT")
    lox2272 = lower_no_whitespace("ATAACTTCGTATA GgATACtT TATACGAAGTTAT")

    loxm2 = lower_no_whitespace("ATAACTTCGTATA TGGTTTCT TATACGAAGTTAT")
    loxm2_66 = lower_no_whitespace("ATAACTTCGTATA TGGTTTCT TATACGAACGGTA")
    loxm2_71 = lower_no_whitespace("TACCGTTCGTATA TGGTTTCT TATACGAAGTTAT")
    loxm2_72 = lower_no_whitespace("TACCGTTCGTATA TGGTTTCT TATACGAACGGTA")


class CreLoxReaction(Reaction):

    @staticmethod
    def allowed():
        return [
            RMCE(Sites.loxP, Sites.lox2272, Sites.loxP, Sites.lox2272, Sites.loxP, Sites.lox2272),
            RMCE(Sites.lox2272, Sites.loxP, Sites.lox2272, Sites.loxP, Sites.lox2272, Sites.loxP),
            RMCE(Sites.lox66, Sites.lox2272,
                 Sites.lox71, Sites.lox2272,
                 Sites.lox72, Sites.lox2272),
            RMCE(Sites.lox2272, Sites.lox66,
                 Sites.lox2272, Sites.lox71,
                 Sites.lox2272, Sites.lox72),
            RMCE(Sites.loxm2_71, Sites.lox66,
                 Sites.loxm2_66, Sites.lox71,
                 Sites.loxP, Sites.lox72),
            RMCE(Sites.loxm2_66, Sites.lox66,
                 Sites.loxm2_71, Sites.lox71,
                 Sites.loxm2_72, Sites.lox72),

            Integration(Sites.lox66, Sites.lox71, Sites.lox72, Sites.loxP),
            Integration(Sites.lox71, Sites.lox66, Sites.loxP, Sites.lox72),

            Excision(Sites.loxP, Sites.loxP, Sites.loxP),
            Excision(Sites.lox66, Sites.loxP, Sites.loxP),
            Excision(Sites.loxP, Sites.lox66, Sites.lox66),
            Excision(Sites.lox71, Sites.loxP, Sites.lox71),
            Excision(Sites.loxP, Sites.lox71, Sites.loxP),

            Inversion(Sites.loxP, rc(Sites.loxP), Sites.loxP, rc(Sites.loxP)),
            Inversion(Sites.lox66, rc(Sites.loxP), Sites.loxP, rc(Sites.lox66)),
            Inversion(Sites.lox71, rc(Sites.loxP), Sites.lox71, rc(Sites.loxP)),
        ]



import os

from django.conf import settings


def make_required_dirs(path):
    dirn = os.path.dirname(path)
    try:
        original_umask = os.umask(0)
        os.makedirs(dirn, 0o777)
    except OSError:
        pass
    finally:
        os.umask(original_umask)


def get_fragment_reference_fasta_gz_fn(fragment_id):
    return f"{settings.SEQUENCE_FILE_DIR}/edge-fragment-{fragment_id}.fa.gz"



from edge.ssr import lower_no_whitespace, rc
from edge.tests.test_ssr_view import SSRTester


Frt = lower_no_whitespace("GAAGTTCCTATTC tctagaaa GTATAGGAACTTC")
Frt3 = lower_no_whitespace("GAAGTTCCTATTC ttcaaata GTATAGGAACTTC")


class FlpTest(SSRTester):

    def test_excise_with_frt(self):
        self.with_genome("t" * 100 + Frt + "aaa" + Frt + "c" * 100) \
            .when_trigger_flp() \
            .modified_sequence_is("t" * 100 + Frt + "c" * 100)

        self.with_genome("t" * 100 + Frt3 + "aaa" + Frt3 + "c" * 100) \
            .when_trigger_flp() \
            .modified_sequence_is("t" * 100 + Frt3 + "c" * 100)

    def test_inversion_with_frt(self):
        self.with_genome("t" * 100 + Frt + "atg" + rc(Frt) + "c" * 100) \
            .when_trigger_flp() \
            .modified_sequence_is("t" * 100 + Frt + "cat" + rc(Frt) + "c" * 100)

        self.with_genome("t" * 100 + Frt3 + "atg" + rc(Frt3) + "c" * 100) \
            .when_trigger_flp() \
            .modified_sequence_is("t" * 100 + Frt3 + "cat" + rc(Frt3) + "c" * 100)

    def test_RMCE_with_frt_frt3(self):
        self.with_genome("t" * 100 + Frt + "atg" + Frt3 + "c" * 100) \
            .when_trigger_flp_with_donor("a" * 10 + Frt + "g" * 10 + Frt3, False) \
            .modified_sequence_is("t" * 100 + Frt + "g" * 10 + Frt3 + "c" * 100)

    def test_RMCE_with_frt3_frt(self):
        self.with_genome("t" * 100 + Frt3 + "atg" + Frt + "c" * 100) \
            .when_trigger_flp_with_donor("a" * 10 + Frt3 + "g" * 10 + Frt, False) \
            .modified_sequence_is("t" * 100 + Frt3 + "g" * 10 + Frt + "c" * 100)



import itertools
from edge.ssr import rc, lower_no_whitespace, Reaction, Excision, Inversion, RMCE


class Sites(object):
    Frt = lower_no_whitespace("GAAGTTCCTATTC tctagaaa GTATAGGAACTTC")
    Frt1 = lower_no_whitespace("GAAGTTCCTATTC tctagata GTATAGGAACTTC")
    Frt2 = lower_no_whitespace("GAAGTTCCTATTC tctactta GTATAGGAACTTC")
    Frt3 = lower_no_whitespace("GAAGTTCCTATTC ttcaaata GTATAGGAACTTC")
    Frt4 = lower_no_whitespace("GAAGTTCCTATTC tctagaag GTATAGGAACTTC")
    Frt5 = lower_no_whitespace("GAAGTTCCTATTC ttcaaaag GTATAGGAACTTC")


class FlpReaction(Reaction):

    @staticmethod
    def allowed():
        frt_sites = [Sites.Frt, Sites.Frt1, Sites.Frt2, Sites.Frt3, Sites.Frt4, Sites.Frt5]

        allowed = []
        for frt in frt_sites:
            allowed.append(Excision(frt, frt, frt))
            allowed.append(Inversion(frt, rc(frt), frt, rc(frt)))

        for combo in list(itertools.combinations(frt_sites, 2)):
            allowed.append(RMCE(combo[0], combo[1], combo[0], combo[1], combo[0], combo[1]))
            allowed.append(RMCE(combo[1], combo[0], combo[1], combo[0], combo[1], combo[0]))

        return allowed



# Generated by Django 2.2.20 on 2022-06-08 20:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edge', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='chunk',
            name='ref_end_index',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='chunk',
            name='ref_start_index',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



from edge.ssr import lower_no_whitespace, rc
from edge.tests.test_ssr_view import SSRTester


loxP = lower_no_whitespace("ATAACTTCGTATA GCATACAT TATACGAAGTTAT")
lox66 = lower_no_whitespace("ATAACTTCGTATA GCATACAT TATACGAACGGTA")
lox71 = lower_no_whitespace("TACCGTTCGTATA GCATACAT TATACGAAGTTAT")
lox72 = lower_no_whitespace("TACCGTTCGTATA GCATACAT TATACGAACGGTA")

lox5171 = lower_no_whitespace("ATAACTTCGTATA GtAcACAT TATACGAAGTTAT")
lox2272 = lower_no_whitespace("ATAACTTCGTATA GgATACtT TATACGAAGTTAT")

loxm2 = lower_no_whitespace("ATAACTTCGTATA TGGTTTCT TATACGAAGTTAT")
loxm2_66 = lower_no_whitespace("ATAACTTCGTATA TGGTTTCT TATACGAACGGTA")
loxm2_71 = lower_no_whitespace("TACCGTTCGTATA TGGTTTCT TATACGAAGTTAT")
loxm2_72 = lower_no_whitespace("TACCGTTCGTATA TGGTTTCT TATACGAACGGTA")


class CreLoxTest(SSRTester):

    def test_integration_with_lox66_onto_lox71(self):
        self.with_genome("t" * 100 + lox71 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox66 + "g" * 10) \
            .modified_sequence_is("t" * 100 + lox72 + "g" * 10 + "a" * 10 + loxP + "c" * 100)

    def test_integration_with_lox66_onto_rc_lox71(self):
        self.with_genome("t" * 100 + rc(lox71) + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox66 + "g" * 10) \
            .modified_sequence_is(
                "t" * 100 + rc(loxP) + "t" * 10 + "c" * 10 + rc(lox72) + "c" * 100
            )

    def test_integration_with_lox71_onto_lox66(self):
        self.with_genome("t" * 100 + lox66 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox71 + "g" * 10) \
            .modified_sequence_is("t" * 100 + loxP + "g" * 10 + "a" * 10 + lox72 + "c" * 100)

    def test_integration_with_lox71_onto_rc_lox66(self):
        self.with_genome("t" * 100 + rc(lox66) + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox71 + "g" * 10) \
            .modified_sequence_is(
                "t" * 100 + rc(lox72) + "t" * 10 + "c" * 10 + rc(loxP) + "c" * 100
            )

    def test_integration_with_lox66_onto_loxP_is_not_allowed_because_immediate_loopout(self):
        self.with_genome("t" * 100 + loxP + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox66 + "g" * 10) \
            .is_not_allowed()

    def test_excision_with_loxP(self):
        self.with_genome("t" * 100 + loxP + "aaa" + loxP + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "c" * 100)

    def test_excision_with_lox66_loxP(self):
        self.with_genome("t" * 100 + lox66 + "aaa" + loxP + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "c" * 100)

    def test_excision_with_loxP_lox66(self):
        self.with_genome("t" * 100 + loxP + "aaa" + lox66 + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + lox66 + "c" * 100)

    def test_excision_with_lox71_loxP(self):
        self.with_genome("t" * 100 + lox71 + "aaa" + loxP + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + lox71 + "c" * 100)

    def test_excision_with_loxP_lox71(self):
        self.with_genome("t" * 100 + loxP + "aaa" + lox71 + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "c" * 100)

    def test_inversion_with_loxP(self):
        self.with_genome("t" * 100 + loxP + "atg" + rc(loxP) + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "cat" + rc(loxP) + "c" * 100)

    def test_inversion_with_loxP_lox66(self):
        self.with_genome("t" * 100 + loxP + "atg" + rc(lox66) + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + lox66 + "cat" + rc(loxP) + "c" * 100)

    def test_inversion_with_loxP_lox71(self):
        self.with_genome("t" * 100 + loxP + "atg" + rc(lox71) + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "cat" + rc(lox71) + "c" * 100)

    def test_inversion_with_lox66_loxP(self):
        self.with_genome("t" * 100 + lox66 + "atg" + rc(loxP) + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + loxP + "cat" + rc(lox66) + "c" * 100)

    def test_inversion_with_lox71_loxP(self):
        self.with_genome("t" * 100 + lox71 + "atg" + rc(loxP) + "c" * 100) \
            .when_trigger_crelox() \
            .modified_sequence_is("t" * 100 + lox71 + "cat" + rc(loxP) + "c" * 100)

    def test_RMCE_with_lox2272_loxP(self):
        self.with_genome("t" * 100 + lox2272 + "atg" + loxP + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox2272 + "g" * 10 + loxP, False) \
            .modified_sequence_is("t" * 100 + lox2272 + "g" * 10 + loxP + "c" * 100)

    def test_RMCE_with_loxP_lox2272(self):
        self.with_genome("t" * 100 + loxP + "atg" + lox2272 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + loxP + "g" * 10 + lox2272, False) \
            .modified_sequence_is("t" * 100 + loxP + "g" * 10 + lox2272 + "c" * 100)

    def test_RMCE_with_lox2272_loxP_and_reversed_on_donor(self):
        self.with_genome("t" * 100 + lox2272 + "atg" + loxP + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + rc(loxP) + "g" * 10 + rc(lox2272), False) \
            .modified_sequence_is("t" * 100 + lox2272 + "c" * 10 + loxP + "c" * 100)

    def test_RMCE_with_lox2272_loxP_and_loxP_lox2272_is_not_allowed(self):
        self.with_genome("t" * 100 + lox2272 + "atg" + loxP + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + loxP + "g" * 10 + lox2272, False) \
            .is_not_allowed()

    def test_RMCE_with_lox71_lox2272_and_lox66_lox2272(self):
        # this also tests that RMCE takes precedence over integration
        self.with_genome("t" * 100 + lox71 + "atg" + lox2272 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox66 + "g" * 10 + lox2272, False) \
            .modified_sequence_is("t" * 100 + lox72 + "g" * 10 + lox2272 + "c" * 100)

    def test_RMCE_with_lox2272_lox71_and_lox2272_lox66(self):
        # this also tests that RMCE takes precedence over integration
        self.with_genome("t" * 100 + lox2272 + "atg" + lox71 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + lox2272 + "g" * 10 + lox66, False) \
            .modified_sequence_is("t" * 100 + lox2272 + "g" * 10 + lox72 + "c" * 100)

    def test_RMCE_with_loxm2_66_lox71_and_loxm2_71_lox66(self):
        # this also tests that RMCE takes precedence over integration
        self.with_genome("t" * 100 + loxm2_66 + "atg" + lox71 + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + loxm2_71 + "g" * 10 + lox66) \
            .modified_sequence_is("t" * 100 + loxP + "g" * 10 + lox72 + "c" * 100)

    def test_RMCE_with_loxP_loxP_and_loxP_loxP_is_not_allowed_because_multiple_recombinations(self):
        self.with_genome("t" * 100 + loxP + "atg" + loxP + "c" * 100) \
            .when_trigger_crelox_with_donor("a" * 10 + loxP + "g" * 10 + loxP) \
            .is_not_allowed()



from Bio.Seq import Seq


# sites separated by more than these number of bps will not be considered as
# part of the same combination
SITE_SEPARATION_MAX = 50000


def rc(s):
    return str(Seq(s).reverse_complement())


def lower_no_whitespace(s):
    return s.replace(" ", "").lower()


def find_indices(big, small):
    """
    Returns 0 based indices of all occurances of small in big.
    """

    res = []
    offset = 0
    while len(big) >= len(small):
        if small in big:
            res.append(offset + big.index(small))
            offset += big.index(small) + len(small)
            big = big[big.index(small) + len(small):]
        else:
            return res
    return res


class SiteLocation(object):

    def __init__(self, site, fragment, insert, is_fragment_or_insert_circular, start_0based):
        self.site = site
        self.fragment = fragment
        self.insert = insert
        self.start_0based = start_0based  # 0-indexed
        self.used = False
        self.is_fragment_or_insert_circular = is_fragment_or_insert_circular

    def use(self):
        self.used = True

    @property
    def fragment_id(self):
        return self.fragment.id if self.fragment is not None else None

    @property
    def on_insert(self):
        return self.insert is not None


class Recombination(object):
    """
    Generic class to model a recombination event.
    """

    def required_genome_sites(self):
        return []

    def required_insert_sites(self):
        return []

    def find_matching_locations(self, all_locations, required_sites, on_insert, errors):
        all_candidate_single_site_locations = \
          [loc for loc in all_locations
           if on_insert == loc.on_insert and loc.site in required_sites and loc.used is False]

        candidate_single_site_locations_by_fragment = {}
        for loc in all_candidate_single_site_locations:
            if loc.fragment_id not in candidate_single_site_locations_by_fragment:
                candidate_single_site_locations_by_fragment[loc.fragment_id] = []
            candidate_single_site_locations_by_fragment[loc.fragment_id].append(loc)

        matching_locations = []

        for fragment_id, candidate_single_site_locations in \
                candidate_single_site_locations_by_fragment.items():

            candidate_single_site_locations = sorted(
                candidate_single_site_locations,
                key=lambda loc: loc.start_0based
            )

            original_candidate_single_site_locations_length = len(candidate_single_site_locations)
            if len(candidate_single_site_locations) > 0 and \
               candidate_single_site_locations[0].is_fragment_or_insert_circular is True:
                candidate_single_site_locations = \
                    candidate_single_site_locations + candidate_single_site_locations

            for i, single_site_loc in enumerate(candidate_single_site_locations):
                if i >= original_candidate_single_site_locations_length:
                    break
                if single_site_loc.site == required_sites[0]:
                    next_locs = candidate_single_site_locations[i:i + len(required_sites)]
                    next_sites = [loc.site for loc in next_locs]
                    site_separation = next_locs[-1].start_0based - next_locs[0].start_0based
                    if next_sites == list(required_sites) and site_separation < SITE_SEPARATION_MAX:

                        has_conflict = False
                        for loc in next_locs:
                            if loc.used:
                                errors.append("Found %s, can trigger multiple events" % loc.site)
                                has_conflict = True
                            loc.use()

                        if not has_conflict:
                            matching_locations.append(next_locs)

        return matching_locations

    def possible_locations(self, all_locations, errors):
        required_insert_sites = self.required_insert_sites()
        insert_locations = []

        if len(required_insert_sites):
            locations_on_insert = []
            reverse_of_required_insert_sites = [rc(s) for s in required_insert_sites][::-1]
            locations_on_insert.extend(
                self.find_matching_locations(all_locations, required_insert_sites, True, errors)
            )
            if required_insert_sites != reverse_of_required_insert_sites:
                locations_on_insert.extend(
                    self.find_matching_locations(
                        all_locations, reverse_of_required_insert_sites, True, errors
                    )
                )

            if len(locations_on_insert) == 0:
                errors.append(
                    "Required site(s) %s on insert missing" % (required_insert_sites,)
                )
                return []
            elif len(locations_on_insert) > 1:
                errors.append(
                    "Requires one site or one set of sites %s on insert, found multiple"
                    % (required_insert_sites,)
                )
                return []

            insert_locations = locations_on_insert[0]

        possible_locations = []

        required_genome_sites = self.required_genome_sites()
        reverse_of_required_genome_sites = [rc(s) for s in required_genome_sites][::-1]
        locations_on_genome = []
        locations_on_genome.extend(
            self.find_matching_locations(all_locations, required_genome_sites, False, errors)
        )
        if required_genome_sites != reverse_of_required_genome_sites:
            locations_on_genome.extend(
                self.find_matching_locations(
                    all_locations, reverse_of_required_genome_sites, False, errors
                )
            )
        for locs in locations_on_genome:
            possible_locations.append(locs + insert_locations)

        return possible_locations

    def generate_events(self, locations, event_cls, errors):
        location_sets = self.possible_locations(locations, errors)
        events = []
        for locs in location_sets:
            events.append(event_cls(self, locs))
        return events


class Integration(Recombination):

    def __init__(self, site_insert, site_genome,
                 recombined_site_left_genome, recombined_site_right_genome):
        super(Integration, self).__init__()
        self.site_insert = site_insert
        self.site_genome = site_genome
        self.recombined_site_left_genome = recombined_site_left_genome
        self.recombined_site_right_genome = recombined_site_right_genome

    def required_genome_sites(self):
        return [self.site_genome]

    def required_insert_sites(self):
        return [self.site_insert]

    def events(self, locations, errors):
        return self.generate_events(locations, IntegrationEvent, errors)


class Excision(Recombination):
    def __init__(self, site_left, site_right, recombined_site):
        super(Excision, self).__init__()
        self.site_left = site_left
        self.site_right = site_right
        self.recombined_site = recombined_site

    def required_genome_sites(self):
        return [self.site_left, self.site_right]

    def required_insert_sites(self):
        return []

    def events(self, locations, errors):
        return self.generate_events(locations, ExcisionEvent, errors)


class Inversion(Recombination):
    def __init__(self, site_left, site_right, recombined_site_left, recombined_site_right):
        super(Inversion, self).__init__()
        assert site_left not in site_right
        assert site_right not in site_left
        self.site_left = site_left
        self.site_right = site_right
        self.recombined_site_left = recombined_site_left
        self.recombined_site_right = recombined_site_right

    def required_genome_sites(self):
        return [self.site_left, self.site_right]

    def required_insert_sites(self):
        return []

    def events(self, locations, errors):
        return self.generate_events(locations, InversionEvent, errors)


class RMCE(Recombination):
    def __init__(self, site_left_insert, site_right_insert, site_left_genome, site_right_genome,
                 recombined_site_left_genome, recombined_site_right_genome):
        super(RMCE, self).__init__()
        assert site_left_insert not in site_right_insert
        assert site_right_insert not in site_left_insert
        assert site_left_genome not in site_right_genome
        assert site_right_genome not in site_left_genome

        self.site_left_insert = site_left_insert
        self.site_right_insert = site_right_insert
        self.site_left_genome = site_left_genome
        self.site_right_genome = site_right_genome
        self.recombined_site_left_genome = recombined_site_left_genome
        self.recombined_site_right_genome = recombined_site_right_genome

    def required_genome_sites(self):
        return [self.site_left_genome, self.site_right_genome]

    def required_insert_sites(self):
        return [self.site_left_insert, self.site_right_insert]

    def events(self, locations, errors):
        return self.generate_events(locations, RMCEEvent, errors)


class Event(object):

    def __init__(self, recombination, locations):
        self.recombination = recombination
        self.locations = locations

    @property
    def fragment_id(self):
        for loc in self.genomic_locations:
            return loc.fragment_id
        assert False

    def is_reversed(self):
        first_loc = self.genomic_locations[0]

        if first_loc.site == self.recombination.required_genome_sites()[0]:
            return False
        if rc(first_loc.site) == self.recombination.required_genome_sites()[-1]:
            return True
        assert False

    @property
    def genomic_locations(self):
        return sorted([l for l in self.locations
                       if l.fragment_id is not None], key=lambda l: l.start_0based)

    @property
    def genomic_start_0based(self):
        return self.genomic_locations[0].start_0based

    def to_dict(self):
        return dict(
            recombination=dict(
                type=self.recombination.__class__.__name__,
                sites=self.recombination.__dict__
            ),
            genomic_locations=[
                dict(
                    site=loc.site,
                    fragment_id=loc.fragment_id,
                    fragment_name=loc.fragment.name,
                    start=loc.start_0based + 1
                ) for loc in self.genomic_locations
            ]
        )


class IntegrationEvent(Event):

    def get_integrated_aligned_with_site_direction(self, insert, is_insert_circular):
        insert = insert * 3
        site_insert = self.recombination.site_insert
        if site_insert not in insert:
            insert = rc(insert)
        assert insert.index(site_insert) >= 0
        bps = find_indices(insert, site_insert)
        assert len(bps) >= 2
        return insert[bps[0] + len(site_insert):bps[1]]

    def run(self, new_fragment, insert, is_insert_circular):
        genomic_locations = self.genomic_locations
        integrated = self.get_integrated_aligned_with_site_direction(insert, is_insert_circular)
        bps_to_replace = len(self.recombination.site_genome)

        if self.is_reversed():
            integrated = rc(integrated)
            new_site_left = rc(self.recombination.recombined_site_right_genome)
            new_site_right = rc(self.recombination.recombined_site_left_genome)
        else:
            new_site_left = self.recombination.recombined_site_left_genome
            new_site_right = self.recombination.recombined_site_right_genome

        integrated = new_site_left + integrated + new_site_right

        new_fragment.replace_bases(
            genomic_locations[0].start_0based + 1,
            bps_to_replace,
            integrated
        )

        return len(integrated) - bps_to_replace


class ExcisionEvent(Event):

    def run(self, new_fragment, insert, is_insert_circular):
        genomic_locations = self.genomic_locations
        assert len(genomic_locations) == 2

        if not self.is_reversed():
            assert genomic_locations[0].site == self.recombination.site_left
            assert genomic_locations[1].site == self.recombination.site_right
            new_site = self.recombination.recombined_site
            bps_to_remove = \
                genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_right)
        else:
            assert genomic_locations[0].site == rc(self.recombination.site_right)
            assert genomic_locations[1].site == rc(self.recombination.site_left)
            new_site = rc(self.recombination.recombined_site)
            bps_to_remove = \
                genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_left)

        new_fragment.replace_bases(
            genomic_locations[0].start_0based + 1,
            bps_to_remove,
            new_site
        )

        return len(new_site) - bps_to_remove


class InversionEvent(Event):

    def run(self, new_fragment, insert, is_insert_circular):
        genomic_locations = self.genomic_locations
        assert len(genomic_locations) == 2

        if not self.is_reversed():
            assert genomic_locations[0].site == self.recombination.site_left
            assert genomic_locations[1].site == self.recombination.site_right
            old_site_left = self.recombination.site_left
            old_site_right = self.recombination.site_right
            new_site_left = self.recombination.recombined_site_left
            new_site_right = self.recombination.recombined_site_right

            bps_to_replace = \
                genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_right)
        else:
            assert genomic_locations[0].site == rc(self.recombination.site_right)
            assert genomic_locations[1].site == rc(self.recombination.site_left)
            old_site_left = rc(self.recombination.site_right)
            old_site_right = rc(self.recombination.site_left)
            new_site_left = rc(self.recombination.recombined_site_right)
            new_site_right = rc(self.recombination.recombined_site_left)

            bps_to_replace = \
                genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_left)

        sequence_to_replace = new_fragment.get_sequence(
            bp_lo=genomic_locations[0].start_0based + len(old_site_left) + 1,
            bp_hi=genomic_locations[1].start_0based + 1 - 1
        )
        new_sequence = new_site_left + rc(sequence_to_replace) + new_site_right
        assert \
            len(old_site_left) + len(sequence_to_replace) + len(old_site_right) == bps_to_replace

        new_fragment.replace_bases(
            genomic_locations[0].start_0based + 1,
            bps_to_replace,
            new_sequence
        )

        return len(new_sequence) - \
            (len(old_site_left) + len(sequence_to_replace) + len(old_site_right))


class RMCEEvent(Event):

    def get_integrated_aligned_with_site_direction(self, insert):
        insert = insert * 2
        site_left_insert = self.recombination.site_left_insert
        site_right_insert = self.recombination.site_right_insert
        if site_left_insert not in insert:
            insert = rc(insert)
        # below, to handle when sites are across circular boundary
        left_trimmed_insert = insert[insert.index(site_left_insert) + len(site_left_insert):]
        return left_trimmed_insert[:left_trimmed_insert.index(site_right_insert)]

    def run(self, new_fragment, insert, is_insert_circular):
        genomic_locations = self.genomic_locations
        integrated = self.get_integrated_aligned_with_site_direction(insert)

        if not self.is_reversed():
            assert genomic_locations[0].site == self.recombination.site_left_genome
            assert genomic_locations[1].site == self.recombination.site_right_genome
            new_site_left = self.recombination.recombined_site_left_genome
            new_site_right = self.recombination.recombined_site_right_genome

            bps_to_replace = genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_right_genome)
        else:
            integrated = rc(integrated)

            assert genomic_locations[0].site == rc(self.recombination.site_right_genome)
            assert genomic_locations[1].site == rc(self.recombination.site_left_genome)
            new_site_left = rc(self.recombination.recombined_site_right_genome)
            new_site_right = rc(self.recombination.recombined_site_left_genome)

            bps_to_replace = genomic_locations[1].start_0based - \
                genomic_locations[0].start_0based + \
                len(self.recombination.site_left_genome)

        integrated = new_site_left + integrated + new_site_right
        new_fragment.replace_bases(
            genomic_locations[0].start_0based + 1,
            bps_to_replace,
            integrated
        )

        return len(integrated) - bps_to_replace


def add_reverse_sites(sites):
    new_sites = []
    for s in sites:
        new_sites.append(s)
        new_sites.append(rc(s))
    return set(new_sites)


def find_site_locations_on_sequence(sequence, is_circular, sites, fragment_obj=None):
    sites = add_reverse_sites(sites)
    template = sequence
    if is_circular:
        template = sequence * 2
    template = template.lower()

    locations = []
    for site in sites:
        locations.extend(
            find_query_locations_on_duplicated_template(
                template, len(sequence), is_circular, site, fragment_obj=fragment_obj
            )
        )

    uniqdict = {(l.fragment_id, l.start_0based):l for l in locations}
    locations = list(uniqdict.values())
    return locations


def find_query_locations_on_duplicated_template(
        template, sequence_len, is_fragment_or_insert_circular, site,
        fragment_obj=None):

    indices = find_indices(template, site)
    indices = [i for i in indices if i < sequence_len]
    return [
      SiteLocation(
          site,
          fragment_obj,
          template[0:sequence_len] if fragment_obj is None else None,
          is_fragment_or_insert_circular,
          i,
      )
      for i in indices
    ]


class Reaction(object):

    @staticmethod
    def allowed():
        return []

    def __init__(self, parent_genome, insert, is_insert_circular, parent_fragments=None):
        self.parent_genome = parent_genome
        if parent_fragments is not None:
            self.parent_fragments = parent_fragments
        else:
            self.parent_fragments = [
                f.indexed_fragment() for f in self.parent_genome.fragments.all()
            ]

        self.insert = insert
        self.is_insert_circular = is_insert_circular
        self.locations = None
        self.events = None
        self.errors = []

    @classmethod
    def __sites_on_genome(cls):
        sites = []
        for recombination in cls.allowed():
            sites.extend(recombination.required_genome_sites())
        return set(sites)

    @classmethod
    def __sites_on_insert(cls):
        sites = []
        for recombination in cls.allowed():
            sites.extend(recombination.required_insert_sites())
        return set(sites)

    def determine_site_locations(self):
        # FIXME this is a very slow implementation, we can make this faster by
        # using cached results

        locations = []

        if self.insert:
            locations.extend(
                find_site_locations_on_sequence(
                    self.insert,
                    self.is_insert_circular,
                    self.__sites_on_insert()
                )
            )

        sites_on_genome = self.__sites_on_genome()
        for fragment in self.parent_fragments:
            locations.extend(
                find_site_locations_on_sequence(
                    fragment.sequence,
                    fragment.circular,
                    sites_on_genome,
                    fragment_obj=fragment
                )
            )

        self.locations = locations

    def group_into_events(self):
        # TODO if we attempt an integration with loxP on loxP, how do we return
        # that as an error not just silently ignore?

        if self.events is None:
            self.determine_site_locations()
            self.events = []
            self.errors = []

            # note that we iterate through allowed() in order - earlier
            # recombination definitions take precedence
            for recombination in self.allowed():
                self.events.extend(recombination.events(self.locations, self.errors))

    def run_reaction(self, new_genome_name, notes=None):
        self.group_into_events()

        if len(self.events) == 0:
            print("errors", self.errors)
            return

        event_insert_sites = [
            site
            for event in self.events
            for site in event.recombination.required_insert_sites()
        ]
        if len(event_insert_sites) == 0 and self.insert:
            print("errors", self.errors)
            print("has insert, but no events use insert")
            return

        new_genome = self.parent_genome.update()
        if new_genome_name:
            new_genome.name = new_genome_name
        else:
            new_genome.name = self.parent_genome.name + " SSR modified"
        new_genome.notes = notes
        new_genome.save()

        old_to_new_fragment_dict = {}

    # needs to run through events by fragment and start bp in reverse order
    # on the fragment, so we don't have to adjust coordinates of events
    # that have not been run yet.

        events = sorted(self.events,
                        key=lambda e: (e.fragment_id, -e.genomic_start_0based))
        while len(events) > 0:
            event = events[0]

            if event.fragment_id not in old_to_new_fragment_dict:
                with new_genome.update_fragment_by_fragment_id(
                    event.fragment_id, new_fragment=True
                ) as f:
                    old_to_new_fragment_dict[event.fragment_id] = f

            new_fragment = old_to_new_fragment_dict[event.fragment_id]
            event.run(new_fragment, self.insert, self.is_insert_circular)

            events = events[1:]

        print("new genome in run_reaction", new_genome)
        return new_genome



import time

from BCBio import GFF
from django.db import connection

from edge.models import Fragment, Fragment_Chunk_Location
from edge.models.chunk import Chunk


GFF_NAME_FIELDS = (
    "name",
    "Name",
    "gene",
    "locus",
    "locus_tag",
    "product",
    "protein_id",
)


def circular_mod(number, seq_length):
    return ((number - 1) % seq_length) + 1


class GFFImporter(object):
    def __init__(self, genome, gff_fasta_fn):
        self.__genome = genome
        self.__gff_fasta_fn = gff_fasta_fn

    def do_import(self, dirn='.'):
        in_file = self.__gff_fasta_fn
        in_handle = open(in_file)

        # In DEBUG=True mode, Django keeps list of queries and blows up memory
        # usage when doing a big import. The following line disables this
        # logging.
        connection.use_debug_cursor = False

        # First, retrieve rec names
        rec_ids = [i[0] for i in GFF.GFFExaminer().available_limits(in_handle).get('gff_id')]
        in_handle.close()

        # Then parse GFF by rec
        t0 = time.time()
        for rec_id in rec_ids:
            in_handle = open(in_file)
            limit_info = dict(gff_id=[rec_id])
            recs = [rec for rec in GFF.parse(in_handle, limit_info=limit_info)
                    if rec.id == rec_id]

            for rec in recs:
                if self.__genome.fragments.filter(name=rec.id).count() > 0:
                    print("skipping %s, already imported" % rec.id)
                    break
                else:
                    try:
                        GFFFragmentImporter(rec, dirn=dirn).parse_gff()
                    except Exception as e:
                        print(str(e))
                        raise Exception(f"{rec} failed import validation")
            in_handle.close()
        print("%s seconds to parse and validate all contigs from GFF" % (time.time() - t0))

        # Then, build and annotate fragments
        for rec_id in rec_ids:
            in_handle = open(in_file)
            limit_info = dict(gff_id=[rec_id])
            recs = [rec for rec in GFF.parse(in_handle, limit_info=limit_info)
                    if rec.id == rec_id]

            for rec in recs:
                importer = GFFFragmentImporter(rec, dirn=dirn)
                fragment = importer.do_import()
                if fragment is None:
                    raise Exception(f"{rec} failed fragment generation")
                self.__genome.genome_fragment_set.create(fragment=fragment, inherited=False)
            in_handle.close()

        # Be nice and turn debug cursor back on
        connection.use_debug_cursor = True


class GFFFragmentImporter(object):
    def __init__(self, gff_rec, dirn='.'):
        self.__rec = gff_rec
        self.__seqlen = len(gff_rec.seq)
        self.__features = None
        self.__fclocs = None
        self.__subfeatures_dict = {}
        self.dirn = dirn

    def finish_import(self):
        t0 = time.time()
        try:
            f = self.build_fragment(dirn=self.dirn)
        except Exception as e:
            print(str(e))
            raise Exception(f"{self.__rec} failed fragment building")
        print("build fragment: %.4f" % (time.time() - t0,))

        t0 = time.time()
        try:
            self.annotate(f)
        except Exception as e:
            print(str(e))
            raise Exception(f"{self.__rec} failed fragment annotation")
        print("annotate: %.4f" % (time.time() - t0,))

        return f

    def do_import(self):
        t0 = time.time()
        try:
            self.parse_gff()
        except Exception as e:
            print(str(e))
            raise Exception(f"{self.__rec} failed import validation")
        print("parse gff: %.4f" % (time.time() - t0,))

        return self.finish_import()

    def parse_gff(self):
        print("%s: %s" % (self.__rec.id, self.__seqlen))

        features = []
        for feature in self.__rec.features:
            # skip features that cover the entire sequence
            if feature.type.upper() in ['REGION', 'CHR', 'CHROM', 'CHROMOSOME']:
                continue

            # get name and qualifiers
            names = list(
                filter(lambda x: len(x) > 0,
                       [feature.qualifiers.get(field, []) for field in GFF_NAME_FIELDS]
                       )
            )
            feature_name = ((names[0][0] if names else None)
                            or feature.id or feature.type or "")[0:100]
            qualifiers = {field: v for field, v in feature.qualifiers.items() if len(v) > 0}

            # start in Genbank format is start after, so +1 here
            features.append(
                (
                    circular_mod(int(feature.location.start) + 1, self.__seqlen),
                    circular_mod(int(feature.location.end), self.__seqlen),
                    feature_name,
                    feature.type,
                    feature.strand,
                    qualifiers,
                )
            )

            # add sub features for chunking for CDS only
            self.__subfeatures_dict[feature_name] = []

            # order sub features based on relative position in the feature
            first, second = [], []
            for sub in sorted(feature.sub_features, key=lambda f: int(f.location.start)):
                if circular_mod(int(sub.location.start) + 1, self.__seqlen) < features[-1][0]:
                    second.append(sub)
                else:
                    first.append(sub)
            sub_feats_to_iter = first + second

            for sub in sub_feats_to_iter:
                # check that the type is right for a sub feature
                if not (sub.type.upper() in ['CDS', 'EXON', 'INTRON']
                        or sub.type.upper()[-3:] == 'RNA'):
                    continue

                # get name and qualifiers
                sub_names = list(
                    filter(lambda x: len(x) > 0,
                           [sub.qualifiers.get(field, []) for field in GFF_NAME_FIELDS]
                           )
                )
                subfeature_name = ((sub_names[0][0] if sub_names else None)
                                   or sub.id or feature_name)[0:100]
                qualifiers = {field: v for field, v in sub.qualifiers.items() if len(v) > 0}

                # start in Genbank format is start after, so +1 here
                sub_tup = (
                        circular_mod(int(sub.location.start) + 1, self.__seqlen),
                        circular_mod(int(sub.location.end), self.__seqlen),
                        subfeature_name,
                        sub.type,
                        sub.strand,
                        qualifiers,
                    )

                # if it has no id, it belongs to the feature
                # otherwise, mark it as its own feature
                if subfeature_name == feature_name:
                    self.__subfeatures_dict[feature_name].append(sub_tup)
                else:
                    if subfeature_name in self.__subfeatures_dict:
                        self.__subfeatures_dict[subfeature_name].append(sub_tup)
                    else:
                        features.append(sub_tup)
                        self.__subfeatures_dict[subfeature_name] = [sub_tup]

                # order sub sub features based on relative position in the sub feature
                first, second = [], []
                for sub_sub in sorted(sub.sub_features, key=lambda f: int(f.location.start)):
                    if circular_mod(int(sub.location.start) + 1, self.__seqlen) < features[-1][0]:
                        second.append(sub_sub)
                    else:
                        first.append(sub_sub)
                sub_sub_feats_to_iter = first + second

                for sub_sub in sub_sub_feats_to_iter:
                    # get name and qualifiers
                    sub_sub_names = list(
                        filter(lambda x: len(x) > 0,
                               [sub_sub.qualifiers.get(field, []) for field in GFF_NAME_FIELDS]
                               )
                    )
                    subsubfeature_name = ((sub_sub_names[0][0] if sub_sub_names else None)
                                          or sub_sub.id or subfeature_name)[0:100]
                    qualifiers = {field: v for field, v in feature.qualifiers.items() if len(v) > 0}

                    # start in Genbank format is start after, so +1 here
                    sub_sub_tup = (
                            circular_mod(int(sub_sub.location.start) + 1, self.__seqlen),
                            circular_mod(int(sub_sub.location.end), self.__seqlen),
                            subsubfeature_name,
                            sub_sub.type,
                            sub_sub.strand,
                            qualifiers,
                        )

                    # if it has no id and the sub feature has no id, it belongs to the feature
                    # if it has no id and the sub feature has id, it belongs to the sub feature
                    # otherwise, mark it as its own feature
                    if subsubfeature_name == feature_name:
                        self.__subfeatures_dict[feature_name].append(sub_sub_tup)
                    elif subsubfeature_name == subfeature_name:
                        self.__subfeatures_dict[subfeature_name].append(sub_sub_tup)
                    else:
                        if subsubfeature_name in self.__subfeatures_dict:
                            self.__subfeatures_dict[subsubfeature_name].append(sub_sub_tup)
                        else:
                            features.append(sub_sub_tup)
                            self.__subfeatures_dict[subsubfeature_name] = [sub_sub_tup]

        self.__features = features

        # update features made from only subfeatures
        for feature in features:
            if self.__subfeatures_dict[feature[2]] != []:
                features.remove(feature)
                new_start = self.__subfeatures_dict[feature[2]][0][0]
                new_end = self.__subfeatures_dict[feature[2]][-1][1]
                new_feature = (new_start, new_end, feature[2], feature[3], feature[4], feature[5])
                features.append(new_feature)

    def build_fragment(self, reference_based=True, dirn='.'):
        # pre-chunk the fragment sequence at feature start and end locations.
        # there should be no need to further divide any chunk during import.
        starts_and_ends = []
        for feature in self.__features:
            name = feature[2]
            starts_and_ends.append(feature[0])
            starts_and_ends.append(feature[1] + 1)
            for subfeature in self.__subfeatures_dict[name]:
                starts_and_ends.append(subfeature[0])
                starts_and_ends.append(subfeature[1] + 1)
        break_points = sorted(list(set(starts_and_ends)))

        cur_len = 0
        chunk_sizes = []
        for i, bp in enumerate(break_points):
            if i == 0:
                if bp > 1:
                    chunk_sizes.append(break_points[i] - 1)
                    cur_len += chunk_sizes[-1]
            else:
                chunk_sizes.append(break_points[i] - break_points[i - 1])
                cur_len += chunk_sizes[-1]

        if cur_len < self.__seqlen:
            chunk_sizes.append(self.__seqlen - cur_len)

        fragment_circular = False
        for feature in self.__rec.features:
            # skip features that cover the entire sequence
            if feature.type.upper() in ['REGION', 'CHR', 'CHROM', 'CHROMOSOME']:
                if 'Is_circular' in feature.qualifiers:
                    fragment_circular = feature.qualifiers['Is_circular'][0].upper() == 'TRUE'
                break

        new_fragment = Fragment(
            name=self.__rec.id, circular=fragment_circular, parent=None, start_chunk=None
        )
        new_fragment.save()
        print("Fragment %s" % (new_fragment.id))
        new_fragment = new_fragment.indexed_fragment()

        if reference_based:
            print("%d chunks" % (len(chunk_sizes),))
            t0 = time.time()
            Chunk.CHUNK_REFERENCE_CLASS.generate_from_fragment(
                new_fragment, str(self.__rec.seq), dirn=dirn
            )
            print("Reference file generation took %s seconds" % (time.time() - t0))

            new_fragment._bulk_create_fragment_chunks(chunk_sizes)
            return new_fragment

        # divide chunks bigger than a certain threshold to smaller chunks, to
        # allow insertion of sequence into database. e.g. MySQL has a packet
        # size that prevents chunks that are too large from being inserted.
        chunk_size_limit = 1000000
        new_chunk_sizes = []
        for original_chunk_size in chunk_sizes:
            if original_chunk_size < chunk_size_limit:
                new_chunk_sizes.append(original_chunk_size)
            else:
                divided_chunks = []
                while original_chunk_size > 0:
                    divided_chunks.append(min(original_chunk_size, chunk_size_limit))
                    original_chunk_size -= chunk_size_limit
                new_chunk_sizes.extend(divided_chunks)
        chunk_sizes = new_chunk_sizes
        print("%d chunks" % (len(chunk_sizes),))

        prev = None
        fragment_len = 0
        for chunk_size in chunk_sizes:
            t0 = time.time()
            prev = new_fragment._append_to_fragment(
                prev,
                fragment_len,
                str(self.__rec.seq[fragment_len : fragment_len + chunk_size]),
            )
            fragment_len += chunk_size
            print("add chunk to fragment: %.4f\r" % (time.time() - t0,), end="")

        return new_fragment

    def annotate(self, fragment):
        self.__fclocs = {
            c.base_first: c
            for c in Fragment_Chunk_Location.objects.select_related("chunk").filter(
                fragment=fragment
            )
        }

        for feature in self.__features:
            t0 = time.time()
            f_start, f_end, f_name, f_type, f_strand, f_qualifiers = feature
            # print('  %s %s: %s-%s %s' % (f_type, f_name, f_start, f_end, f_strand))
            if self.__subfeatures_dict[f_name] != []:
                f_qualifiers['subfeature_qualifiers'] = {
                    f"{sf[0]}_{sf[1]}": sf[5] for sf in self.__subfeatures_dict[f_name]
                    if sf[0] != f_start and sf[1] != f_end
                }
            new_feature = self._annotate_feature(
                fragment, f_start, f_end, f_name, f_type, f_strand, f_qualifiers
            )
            if new_feature is None:
                continue
            feature_base_first = 1
            sorted_subfeatures = self.__subfeatures_dict[f_name]
            if f_strand == -1:
                sorted_subfeatures.reverse()
            for subfeature in sorted_subfeatures:
                sf_start, sf_end, sf_name, sf_type, sf_strand, sf_qualifiers = subfeature
                self._annotate_feature(
                    fragment, sf_start, sf_end, sf_name, sf_type, sf_strand, sf_qualifiers,
                    feature=new_feature, feature_base_first=feature_base_first
                )
                feature_base_first += sf_end - sf_start + 1
            print("annotate feature: %.4f\r" % (time.time() - t0,), end="")

    def _annotate_feature(
        self, fragment, first_base1, last_base1, name, type, strand, qualifiers,
        feature=None, feature_base_first=1
    ):
        wrap_around = fragment.circular and last_base1 < first_base1
        if wrap_around:
            # has to figure out the total length from last chunk
            length = self.__seqlen - first_base1 + 1 + last_base1
        else:
            length = last_base1 - first_base1 + 1
            if length <= 0:
                raise Exception("Annotation must have length one or more")

        if first_base1 not in self.__fclocs or (
            (last_base1 < self.__seqlen and last_base1 + 1 not in self.__fclocs)) or (
            last_base1 > self.__seqlen
        ):
            """
            raise Exception(
                "Cannot find appropriate sequence for feature: %s, start %s, end %s"
                % (name, first_base1, last_base1)
            )
            """
            # 11/03/2021 - ignoring features at end of GFF that went beyond the last bp
            return None

        bases = []
        for key in self.__fclocs:
            fcloc = self.__fclocs[key]
            if fcloc.base_first >= first_base1 and fcloc.base_last <= last_base1:
                bases.append((fcloc.base_first, fcloc.base_last))
            elif wrap_around:
                if fcloc.base_first >= first_base1 or fcloc.base_last <= last_base1:
                    bases.append((fcloc.base_first, fcloc.base_last))
        sorting_strand = strand if strand is not None else 1
        bases.sort(key=lambda x: (wrap_around * (x[1] <= last_base1), sorting_strand * x[0]))

        length = 0
        for first_base1, last_base1 in bases:
            length += fragment.bp_covered_length(first_base1, last_base1)

        new_feature = fragment._add_feature(
            name, type, length, strand, qualifiers
        ) if feature is None else feature

        if feature is not None or self.__subfeatures_dict[name] == []:
            for first_base1, last_base1 in bases:
                region_length = fragment.bp_covered_length(first_base1, last_base1)
                fragment.annotate_chunk(
                    new_feature, feature_base_first, first_base1, last_base1
                )
                feature_base_first += region_length

        return new_feature



from django.test import TestCase

from edge.models import Genome, Fragment, Genome_Fragment
from edge.ssr import (
  rc,
  find_query_locations_on_duplicated_template,
  find_site_locations_on_sequence,
  Event,
  Excision,
  Inversion,
  Integration,
  RMCE,
  SiteLocation,
  Recombination,
  Reaction,
  SITE_SEPARATION_MAX as SSMAX
)


class FindQueryOnSequenceTest(TestCase):
    def test_finds_locations_on_sequence(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100

        locs = find_query_locations_on_duplicated_template(template, len(template), False, site)
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site)
        self.assertEquals(locs[0].start_0based, 100)
        self.assertEquals(locs[0].is_fragment_or_insert_circular, False)

    def test_finds_locations_on_circular_sequence_and_remember_circularity_on_location_object(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100

        locs = find_query_locations_on_duplicated_template(template, len(template), True, site)
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site)
        self.assertEquals(locs[0].start_0based, 100)
        self.assertEquals(locs[0].is_fragment_or_insert_circular, True)

    def test_mark_location_as_on_fragment(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100

        locs = find_query_locations_on_duplicated_template(
            template, len(template), False, site, fragment_obj=object()
        )
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, False)

    def test_finds_multiple_locations(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100 + site + "c" * 100

        locs = find_query_locations_on_duplicated_template(template, len(template), False, site)
        self.assertEquals(len(locs), 2)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site)
        self.assertEquals(locs[0].start_0based, 100)
        self.assertEquals(locs[1].on_insert, True)
        self.assertEquals(locs[1].site, site)
        self.assertEquals(locs[1].start_0based, 204)

    def test_ignores_starting_position_after_sequence_length_argument(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100 + site + "c" * 100

        locs = find_query_locations_on_duplicated_template(template, len(template), False, site)
        self.assertEquals(len(locs), 2)

        locs = find_query_locations_on_duplicated_template(template, 102, False, site)
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site)
        self.assertEquals(locs[0].start_0based, 100)


class FindLocationsTest(TestCase):
    def test_finds_locations_on_sequence(self):
        site = "aggc"
        template = "a" * 100 + site + "g" * 100

        locs = find_site_locations_on_sequence(template, False, [site])
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site)
        self.assertEquals(locs[0].start_0based, 100)

    def test_finds_multiple_sites_on_sequence(self):
        site1 = "aggc"
        site2 = "tggc"
        template = "a" * 100 + site1 + "g" * 100 + site2 + "t" * 100

        locs = find_site_locations_on_sequence(template, False, [site1, site2])
        self.assertEquals(len(locs), 2)
        locs = sorted(locs, key=lambda l: l.site)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site1)
        self.assertEquals(locs[0].start_0based, 100)
        self.assertEquals(locs[1].on_insert, True)
        self.assertEquals(locs[1].site, site2)
        self.assertEquals(locs[1].start_0based, 204)

    def test_finds_reverse_complement_locations_on_sequence(self):
        site1 = "aggc"
        site2 = "tggc"
        self.assertEquals(rc(site2), "gcca")
        template = "a" * 100 + site1 + "g" * 100 + rc(site2) + "t" * 100

        locs = find_site_locations_on_sequence(template, False, [site1, site2])
        self.assertEquals(len(locs), 2)
        locs = sorted(locs, key=lambda l: l.site)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site1)
        self.assertEquals(locs[0].start_0based, 100)
        self.assertEquals(locs[1].on_insert, True)
        self.assertEquals(locs[1].site, rc(site2))
        self.assertEquals(locs[1].start_0based, 204)

    def test_finds_forward_location_across_circular_boundary(self):
        site1 = "aggc"
        template = site1[2:] + "g" * 100 + site1[:2]

        # cannot find anything if sequence is not circular
        locs = find_site_locations_on_sequence(template, False, [site1])
        self.assertEquals(len(locs), 0)

        locs = find_site_locations_on_sequence(template, True, [site1])
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, site1)
        self.assertEquals(locs[0].start_0based, 102)

    def test_finds_reverse_location_across_circular_boundary(self):
        site1 = "aggc"
        template = rc(site1)[2:] + "g" * 100 + rc(site1)[:2]

        # cannot find anything if sequence is not circular
        locs = find_site_locations_on_sequence(template, False, [site1])
        self.assertEquals(len(locs), 0)

        locs = find_site_locations_on_sequence(template, True, [site1])
        self.assertEquals(len(locs), 1)
        self.assertEquals(locs[0].on_insert, True)
        self.assertEquals(locs[0].site, rc(site1))
        self.assertEquals(locs[0].start_0based, 102)


class ReactionTest(TestCase):

    def test_gathers_locations_on_genome(self):
        class FakeFragment(object):
            def __init__(self, n, s, c):
                self.id = n
                self.sequence = s
                self.circular = c

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [FakeRecombination()]

        f = FakeFragment(14, "c" * 100 + "aggc" + "t" * 100, False)
        r = FakeReaction(None, None, None, [f])
        r.determine_site_locations()

        self.assertEquals(len(r.locations), 1)
        self.assertEquals(r.locations[0].on_insert, False)
        self.assertEquals(r.locations[0].fragment_id, 14)
        self.assertEquals(r.locations[0].site, "aggc")
        self.assertEquals(r.locations[0].start_0based, 100)

    def test_gathers_locations_on_multiple_fragments(self):
        class FakeFragment(object):
            def __init__(self, n, s, c):
                self.id = n
                self.sequence = s
                self.circular = c

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [FakeRecombination()]

        f1 = FakeFragment(14, "c" * 100 + "aggc" + "t" * 100, False)
        f2 = FakeFragment(15, "c" * 33 + "aggc" + "t" * 100, False)
        r = FakeReaction(None, None, None, [f1, f2])
        r.determine_site_locations()

        self.assertEquals(len(r.locations), 2)
        self.assertEquals(r.locations[0].on_insert, False)
        self.assertEquals(r.locations[0].fragment_id, 14)
        self.assertEquals(r.locations[0].site, "aggc")
        self.assertEquals(r.locations[0].start_0based, 100)
        self.assertEquals(r.locations[1].on_insert, False)
        self.assertEquals(r.locations[1].fragment_id, 15)
        self.assertEquals(r.locations[1].site, "aggc")
        self.assertEquals(r.locations[1].start_0based, 33)

    def test_gathers_locations_on_insert_and_genome(self):
        class FakeFragment(object):
            def __init__(self, n, s, c):
                self.id = n
                self.sequence = s
                self.circular = c

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

            def required_insert_sites(self):
                return ["ttag"]

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [FakeRecombination()]

        f = FakeFragment(14, "c" * 100 + "aggc" + "t" * 100, False)
        r = FakeReaction(None, "g" * 33 + "ttag" + "c" * 100, True, [f])
        r.determine_site_locations()

        self.assertEquals(len(r.locations), 2)
        self.assertEquals(r.locations[0].on_insert, True)
        self.assertEquals(r.locations[0].fragment_id, None)
        self.assertEquals(r.locations[0].site, "ttag")
        self.assertEquals(r.locations[0].start_0based, 33)
        self.assertEquals(r.locations[1].on_insert, False)
        self.assertEquals(r.locations[1].fragment_id, 14)
        self.assertEquals(r.locations[1].site, "aggc")
        self.assertEquals(r.locations[1].start_0based, 100)

    def test_group_locations_into_multiple_events(self):
        class FakeFragment(object):
            def __init__(self, n, s, c):
                self.id = n
                self.sequence = s
                self.circular = c

        class FakeEvent(Event):
            pass

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

            def required_insert_sites(self):
                return ["ttag"]

            def events(self, locations, errors):
                return self.generate_events(locations, FakeEvent, errors)

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [FakeRecombination()]

        f = FakeFragment(14, "c" * 100 + "aggc" + "t" * 100 + "aggc" + "g" * 100, False)
        r = FakeReaction(None, "g" * 33 + "ttag" + "c" * 100, True, [f])
        r.group_into_events()

        self.assertEquals(len(r.events), 2)

        self.assertEquals(type(r.events[0]), FakeEvent)
        locations = sorted(r.events[0].locations, key=lambda l: l.on_insert)
        self.assertEquals(len(locations), 2)
        self.assertEquals(locations[0].on_insert, False)
        self.assertEquals(locations[0].fragment_id, 14)
        self.assertEquals(locations[0].site, "aggc")
        self.assertEquals(locations[0].start_0based, 100)
        self.assertEquals(locations[1].on_insert, True)
        self.assertEquals(locations[1].fragment_id, None)
        self.assertEquals(locations[1].site, "ttag")
        self.assertEquals(locations[1].start_0based, 33)

        self.assertEquals(type(r.events[1]), FakeEvent)
        locations = sorted(r.events[1].locations, key=lambda l: l.on_insert)
        self.assertEquals(len(locations), 2)
        self.assertEquals(locations[0].on_insert, False)
        self.assertEquals(locations[0].fragment_id, 14)
        self.assertEquals(locations[0].site, "aggc")
        self.assertEquals(locations[0].start_0based, 204)
        self.assertEquals(locations[1].on_insert, True)
        self.assertEquals(locations[1].fragment_id, None)
        self.assertEquals(locations[1].site, "ttag")
        self.assertEquals(locations[1].start_0based, 33)

    def test_prioritize_based_on_allowed_array_order(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "attg" + "t" * 1000 + "attc" + "c" * 100
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        recomb_rmce = RMCE("attg", "attc", "attg", "attc", "gggg", "gggg")
        recomb_inte = Integration("attg", "attg", "cccc", "cccc")

        class FakeReaction1(Reaction):
            @staticmethod
            def allowed():
                return [recomb_inte]

        class FakeReaction2(Reaction):
            @staticmethod
            def allowed():
                return [recomb_rmce, recomb_inte]

        donor = "ggg" + "attg" + "a" * 100 + "attc"

        # integration only
        r = FakeReaction1(parent_genome, donor, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "cccc" + "a" * 100 + "attc" + "ggg" + "cccc" + "t" * 1000 + "attc" + "c" * 100
        )

        # RMCE takes precedence because it's defined earlier
        r = FakeReaction2(parent_genome, donor, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "gggg" + "a" * 100 + "gggg" + "c" * 100)


class RecombinationTest(TestCase):

    def test_recognizes_matched_locations_on_genome(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 1)
        self.assertEquals(matched[0][0].site, "aggc")

    def test_recognizes_matched_locations_in_reverse(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["gcct"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 1)
        self.assertEquals(matched[0][0].site, "aggc")

    def test_recognizes_multiple_matched_locations(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("gcct", FakeFragment(11), None, False, 16),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 2)
        self.assertEquals(len(matched[0]), 1)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(len(matched[1]), 1)
        self.assertEquals(matched[1][0].site, "gcct")

    def test_recognizes_site_combinations(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat", "gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("gaat", FakeFragment(11), None, False, 26),
            SiteLocation("gggc", FakeFragment(11), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(matched[0][0].start_0based, 12)
        self.assertEquals(matched[0][1].site, "gaat")
        self.assertEquals(matched[0][1].start_0based, 26)
        self.assertEquals(matched[0][2].site, "gggc")
        self.assertEquals(matched[0][2].start_0based, 88)

    def test_recognizes_site_combinations_in_reverse(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat", "gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("gccc", FakeFragment(11), None, False, 12),
            SiteLocation("attc", FakeFragment(11), None, False, 26),
            SiteLocation("gcct", FakeFragment(11), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "gccc")
        self.assertEquals(matched[0][0].start_0based, 12)
        self.assertEquals(matched[0][1].site, "attc")
        self.assertEquals(matched[0][1].start_0based, 26)
        self.assertEquals(matched[0][2].site, "gcct")
        self.assertEquals(matched[0][2].start_0based, 88)

    def test_does_not_recognize_combination_where_sites_are_reversed_but_in_forward_order(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat", "gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("gcct", FakeFragment(11), None, False, 12),
            SiteLocation("attc", FakeFragment(11), None, False, 26),
            SiteLocation("gccc", FakeFragment(11), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 0)

    def test_recognizes_required_site_on_insert(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat"]

            def required_insert_sites(self):
                return ["gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("gggc", None, object(), False, 12),
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("attc", FakeFragment(11), None, False, 26),
            SiteLocation("gcct", FakeFragment(11), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "attc")
        self.assertEquals(matched[0][0].start_0based, 26)
        self.assertEquals(matched[0][0].on_insert, False)
        self.assertEquals(matched[0][1].site, "gcct")
        self.assertEquals(matched[0][1].start_0based, 88)
        self.assertEquals(matched[0][1].on_insert, False)
        self.assertEquals(matched[0][2].site, "gggc")
        self.assertEquals(matched[0][2].start_0based, 12)
        self.assertEquals(matched[0][2].on_insert, True)

    def test_recognizes_required_site_in_reverse_on_insert(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat"]

            def required_insert_sites(self):
                return ["gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("gccc", None, object(), False, 12),
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("attc", FakeFragment(11), None, False, 26),
            SiteLocation("gcct", FakeFragment(11), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "attc")
        self.assertEquals(matched[0][0].start_0based, 26)
        self.assertEquals(matched[0][0].on_insert, False)
        self.assertEquals(matched[0][1].site, "gcct")
        self.assertEquals(matched[0][1].start_0based, 88)
        self.assertEquals(matched[0][1].on_insert, False)
        self.assertEquals(matched[0][2].site, "gccc")
        self.assertEquals(matched[0][2].start_0based, 12)
        self.assertEquals(matched[0][2].on_insert, True)

    def test_recognizes_required_site_on_insert_across_circular_boundary(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc"]

            def required_insert_sites(self):
                return ["gggc", "ggga"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("gggc", None, object(), True, 12),
            SiteLocation("ggga", None, object(), True, 2),
            SiteLocation("aggc", FakeFragment(11), None, False, 33),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 1)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(matched[0][0].start_0based, 33)
        self.assertEquals(matched[0][0].on_insert, False)
        self.assertEquals(matched[0][1].site, "gggc")
        self.assertEquals(matched[0][1].start_0based, 12)
        self.assertEquals(matched[0][1].on_insert, True)
        self.assertEquals(matched[0][2].site, "ggga")
        self.assertEquals(matched[0][2].start_0based, 2)
        self.assertEquals(matched[0][2].on_insert, True)

    def test_does_not_recognize_sites_across_chromosomes_as_one_set(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat", "gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("gaat", FakeFragment(11), None, False, 26),
            SiteLocation("gggc", FakeFragment(12), None, False, 88),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 0)

    def test_does_recognize_different_combinations_of_sites_on_different_chromosomes(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "gaat", "gggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("atgc", FakeFragment(11), None, False, 8),
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("gaat", FakeFragment(11), None, False, 26),
            SiteLocation("gggc", FakeFragment(11), None, False, 88),
            SiteLocation("aggc", FakeFragment(12), None, False, 42),
            SiteLocation("gaat", FakeFragment(12), None, False, 56),
            SiteLocation("gggc", FakeFragment(12), None, False, 78),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 2)
        self.assertEquals(len(matched[0]), 3)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(matched[0][0].start_0based, 12)
        self.assertEquals(matched[0][0].fragment_id, 11)
        self.assertEquals(len(matched[1]), 3)
        self.assertEquals(matched[1][0].site, "aggc")
        self.assertEquals(matched[1][0].start_0based, 42)
        self.assertEquals(matched[1][0].fragment_id, 12)

    def test_ignores_sites_too_far_part_and_considers_those_that_are_closer_together(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "aggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("aggc", FakeFragment(11), None, False, 12 + SSMAX + 1),
            SiteLocation("aggc", FakeFragment(11), None, False, 12 + SSMAX + 50),
            SiteLocation("aggc", FakeFragment(11), None, False, 12 + SSMAX + 50 + SSMAX + 1),
            SiteLocation("aggc", FakeFragment(11), None, False,
                         12 + SSMAX + 50 + SSMAX + 1 + SSMAX + 1),
            SiteLocation("aggc", FakeFragment(12), None, False, 42),
            SiteLocation("aggc", FakeFragment(12), None, False, 78),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        self.assertEquals(len(matched), 2)
        self.assertEquals(len(matched[0]), 2)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(matched[0][0].start_0based, 12 + SSMAX + 1)
        self.assertEquals(matched[0][0].fragment_id, 11)
        self.assertEquals(matched[0][1].site, "aggc")
        self.assertEquals(matched[0][1].start_0based, 12 + SSMAX + 50)
        self.assertEquals(matched[0][1].fragment_id, 11)
        self.assertEquals(len(matched[1]), 2)
        self.assertEquals(matched[1][0].site, "aggc")
        self.assertEquals(matched[1][0].start_0based, 42)
        self.assertEquals(matched[1][0].fragment_id, 12)
        self.assertEquals(matched[1][1].site, "aggc")
        self.assertEquals(matched[1][1].start_0based, 78)
        self.assertEquals(matched[1][1].fragment_id, 12)

    def test_ignores_site_combinations_with_multiple_events(self):

        class FakeRecombination(Recombination):
            def required_genome_sites(self):
                return ["aggc", "aggc"]

        class FakeFragment(object):
            def __init__(self, n):
                self.id = n

        locations = [
            SiteLocation("aggc", FakeFragment(11), None, False, 12),
            SiteLocation("aggc", FakeFragment(11), None, False, 22),
            SiteLocation("aggc", FakeFragment(11), None, False, 33),
            SiteLocation("aggc", FakeFragment(12), None, False, 42),
            SiteLocation("aggc", FakeFragment(12), None, False, 78),
        ]

        matched = FakeRecombination().possible_locations(locations, [])
        # 22 and 33 combo is ignored
        self.assertEquals(len(matched), 2)
        self.assertEquals(len(matched[0]), 2)
        self.assertEquals(matched[0][0].site, "aggc")
        self.assertEquals(matched[0][0].start_0based, 12)
        self.assertEquals(matched[0][0].fragment_id, 11)
        self.assertEquals(matched[0][1].site, "aggc")
        self.assertEquals(matched[0][1].start_0based, 22)
        self.assertEquals(matched[0][1].fragment_id, 11)
        self.assertEquals(len(matched[1]), 2)
        self.assertEquals(matched[1][0].site, "aggc")
        self.assertEquals(matched[1][0].start_0based, 42)
        self.assertEquals(matched[1][0].fragment_id, 12)
        self.assertEquals(matched[1][1].site, "aggc")
        self.assertEquals(matched[1][1].start_0based, 78)
        self.assertEquals(matched[1][1].fragment_id, 12)


class ExcisionRecombinationTest(TestCase):
    def test_can_excise_multiple_places_on_fragment(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        seed_sequence = "attg" + "t" * 100 + "cttg" + "c" * 1000 + "attg" + "g" * 100 +\
            "cttg" + "g" * 1000 + "attg" + "a" * 100 + "cttg"
        f = Fragment.create_with_sequence("bar", seed_sequence)

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Excision("attg", "cttg", "attc")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "attc" + "c" * 1000 + "attc" + "g" * 1000 + "attc")

    def test_can_excise_reverse_sites(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "cttg" + "t" + "attg" + "c" * 1000 + "caag" + "g" * 100 + "caat"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Excision("attg", "cttg", "attc")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "cttg" + "t" + "attg" + "c" * 1000 + "gaat")

    def test_works_with_left_and_right_sites_of_different_length(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "cttga" + "t" + "att" + "c" * 1000 + "acaag" + "g" * 100 + "caat"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Excision("attg", "cttgt", "att"),
                        Excision("cttga", "att", "ggg")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "ggg" + "c" * 1000 + "aat")


class InversionRecombinationTest(TestCase):

    def test_can_invert_using_same_site(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "attg" + "atg" + "caat" + "c" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Inversion("attg", "caat", "attg", "caat")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "attg" + "cat" + "caat" + "c" * 1000)

    def test_can_invert_multiple_places_on_fragment(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        seed_sequence = "attg" + "t" * 100 + "cttg" + "c" * 1000 + "attg" +\
            "g" * 100 + "cttg" + "g" * 1000 + "attg" + "a" * 100 + "cttg"
        f = Fragment.create_with_sequence("bar", seed_sequence)

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Inversion("attg", "cttg", "aatt", "ccgg")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()

        new_sequence = "aatt" + "a" * 100 + "ccgg" + "c" * 1000 + "aatt" +\
            "c" * 100 + "ccgg" + "g" * 1000 + "aatt" + "t" * 100 + "ccgg"
        self.assertEquals(f.sequence, new_sequence)

    def test_can_invert_reverse_sites(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "cttg" + "t" + "attg" + "c" * 1000 + "caag" + "g" * 100 + "caat"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Inversion("attg", "cttg", "atta", "cggc")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "cttg" + "t" + "attg" + "c" * 1000 + "gccg" + "c" * 100 + "taat"
        )

    def test_works_with_left_and_right_sites_of_different_length(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "cttga" + "t" + "att" + "c" * 1000 + "acaag" + "g" * 100 + "caat"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Inversion("attg", "cttgt", "att", "gcccgg"),
                        Inversion("cttga", "att", "gg", "cccccc")]

        r = FakeReaction(parent_genome, None, None)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "gg" + "a" + "cccccc" + "c" * 1000 + "ccgggc" + "c" * 100 + "aat"
        )


class IntegrationRecombinationTest(TestCase):

    def test_can_integrate_multiple_places_on_fragment(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "attg" + "t" * 1000 + "attg" + "c" * 100
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("attg", "attg", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()

        new_sequence = "attc" + "a" * 100 + "ggg" + "atcc" + "t" * 1000 + "attc" +\
            "a" * 100 + "ggg" + "atcc" + "c" * 100
        self.assertEquals(f.sequence, new_sequence)

    def test_can_integrate_reverse_sites_on_genome(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "attg" + "t" * 1000 + "caat" + "c" * 100
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("attg", "caat", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()

        new_sequence = "ggat" + "ccc" + "t" * 100 + "gaat" + "t" * 1000 +\
            "attc" + "a" * 100 + "ggg" + "atcc" + "c" * 100
        self.assertEquals(f.sequence, new_sequence)

    def test_can_integrate_reverse_sites_on_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "attg" + "t" * 1000 + "attg" + "c" * 100
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("attg", "attg", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ggg" + "caat" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()

        new_sequence = "attc" + "ccc" + "t" * 100 + "atcc" + "t" * 1000 +\
            "attc" + "ccc" + "t" * 100 + "atcc" + "c" * 100
        self.assertEquals(f.sequence, new_sequence)

    def test_can_integrate_reverse_sites_on_genome_and_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 1000 + "attg" + "c" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("caat", "caat", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "t" * 1000 + "ggat" + "a" * 100 + "ggg" + "gaat" + "c" * 1000
        )

    def test_works_with_left_and_right_sites_of_different_length(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 1000 + "attg" + "c" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("cttg", "attg", "aattc", "atcccc")]

        r = FakeReaction(parent_genome, "ggg" + "cttg" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "t" * 1000 + "aattc" + "a" * 100 + "ggg" + "atcccc" + "c" * 1000
        )

    def test_can_integrate_insert_with_different_site(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 1000 + "attg" + "c" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("cttg", "attg", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ggg" + "cttg" + "a" * 100, True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "t" * 1000 + "attc" + "a" * 100 + "ggg" + "atcc" + "c" * 1000
        )

    def test_can_integrate_insert_with_site_across_circular_boundary(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "t" * 1000 + "attg" + "c" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [Integration("cttg", "attg", "attc", "atcc")]

        r = FakeReaction(parent_genome, "ttg" + "a" * 100 + "c", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "t" * 1000 + "attc" + "a" * 100 + "atcc" + "c" * 1000
        )


class RMCERecombinationTest(TestCase):

    def test_can_integrate_multiple_places_on_fragment(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "ggaa" + "c" * 100 + "gcaa" + "t" * 1000 + "ggaa" + "c" * 100 + "gcaa"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attc", "ggaa", "gcaa", "cctt", "ccta")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100 + "attc", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "cctt" + "a" * 100 + "ccta" + "t" * 1000 + "cctt" + "a" * 100 + "ccta"
        )

    def test_can_integrate_reverse_sites_on_genome(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "ttgc" + "c" * 100 + "ttcc" + "t" * 1000 + "ggaa" + "c" * 100 + "gcaa"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attc", "ggaa", "gcaa", "ccta", "cctt")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100 + "attc", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "aagg" + "t" * 100 + "tagg" + "t" * 1000 + "ccta" + "a" * 100 + "cctt"
        )

    def test_can_integrate_reverse_sites_on_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "ggaa" + "c" * 100 + "gcaa" + "t" * 1000 + "ggaa" + "c" * 100 + "gcaa"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attc", "ggaa", "gcaa", "ccta", "cctt")]

        r = FakeReaction(parent_genome, "ggg" + "gaat" + "a" * 100 + "caat", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(
            f.sequence,
            "ccta" + "t" * 100 + "cctt" + "t" * 1000 + "ccta" + "t" * 100 + "cctt"
        )

    def test_can_integrate_reverse_sites_on_genome_and_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "a" + "ttgc" + "c" * 100 + "ttcc"
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attc", "ggaa", "gcaa", "cctt", "ccta")]

        r = FakeReaction(parent_genome, "ggg" + "gaat" + "a" * 100 + "caat", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "a" + "tagg" + "a" * 100 + "aagg")

    def test_works_with_left_and_right_sites_of_different_length(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "gaa" + "c" * 100 + "gcaa" + "t" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attcc", "gaa", "gcaa", "cctta", "cctaaa")]

        r = FakeReaction(parent_genome, "ggg" + "attg" + "a" * 100 + "attcc", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "cctta" + "a" * 100 + "cctaaa" + "t" * 1000)

    def test_can_integrate_with_sites_across_circular_boundary_on_insert(self):
        parent_genome = Genome(name="foo")
        parent_genome.save()
        f = Fragment.create_with_sequence(
            "bar",
            "ggaa" + "c" * 100 + "gcaa" + "t" * 1000
        )

        Genome_Fragment(genome=parent_genome, fragment=f, inherited=False).save()

        class FakeReaction(Reaction):
            @staticmethod
            def allowed():
                return [RMCE("attg", "attc", "ggaa", "gcaa", "cctt", "ccta")]

        r = FakeReaction(parent_genome, "g" + "a" * 10 + "attc" + "ggg" + "att", True)
        child_genome = r.run_reaction("far")
        f = child_genome.fragments.all()[0].indexed_fragment()
        self.assertEquals(f.sequence, "cctt" + "a" * 10 + "ccta" + "t" * 1000)



import json
from edge.models import Operation
from edge.ssr.crelox import CreLoxReaction
from edge.ssr.flp import FlpReaction


class SSROp(object):
    JSON_ARGUMENT_REACTION_CRELOX = "crelox"
    JSON_ARGUMENT_REACTION_FLP = "flp"

    SUPPORTED = [
        JSON_ARGUMENT_REACTION_CRELOX,
        JSON_ARGUMENT_REACTION_FLP
    ]

    @staticmethod
    def get_reaction(
        genome,
        donor,
        is_donor_circular,
        reaction_name
    ):
        if reaction_name.lower() not in SSROp.SUPPORTED:
            raise Exception("Unknown SSR reaction, expecting one of %s" % (SSROp.SUPPORTED,))

        if reaction_name.lower() == SSROp.JSON_ARGUMENT_REACTION_CRELOX:
            reaction = CreLoxReaction(genome, donor, is_donor_circular)
        elif reaction_name.lower() == SSROp.JSON_ARGUMENT_REACTION_FLP:
            reaction = FlpReaction(genome, donor, is_donor_circular)
        else:
            assert False

        return reaction

    @staticmethod
    def check(
        genome,
        donor,
        is_donor_circular,
        reaction_name,
        **kwargs
    ):
        reaction = SSROp.get_reaction(genome, donor, is_donor_circular, reaction_name)
        reaction.group_into_events()
        return reaction.events

    @staticmethod
    def get_operation(
        donor,
        is_donor_circular,
        reaction_name,
        **kwargs
    ):
        params = dict(donor=donor, is_donor_circular=is_donor_circular, reaction_name=reaction_name)
        op = Operation(type=Operation.SSR[0], params=json.dumps(params))
        return op

    @staticmethod
    def perform(
        genome,
        donor,
        is_donor_circular,
        reaction_name,
        genome_name,
        notes,
        annotations=None,
    ):
        reaction = SSROp.get_reaction(genome, donor, is_donor_circular, reaction_name)
        new_genome = reaction.run_reaction(genome_name, notes=notes)
        return new_genome



from edge.ssr import lower_no_whitespace, Reaction


# TODO:
#
# See original SSR Edge dev plan
# https://docs.google.com/document/d/15iXSn6uWsn16v7nkkxzP9r2korEwr39G_sRlYKbX6IY/edit?pli=1
#
# 1. Add attL and attR sites
# 2. Add rules for integration and RMCE, allowing attB_rc(attP) as well as attB_attP
# 3. Add alternative/orthogonal attB/P sites
# 4. Add separate PhiC31 and Bxb1 reactions
# 5. Figure out how often they would be used in combo, and if we need combined reaction


class Sites(object):
    """
    PhiC31 sites
    """

    PhiC31_attB_tt = lower_no_whitespace("TGCGGGTGCCAGGGCGTGCCC tt GGGCTCCCCGGGCGCGTACTCC")
    PhiC31_attP_tt = lower_no_whitespace("GTGCCCCAACTGGGGTAACCT tt GAGTTCTCTCAGTTGGGGG")

    """
    Bxb1 sites
    """

    Bxb1_attB = lower_no_whitespace("TCGGCCGGCTTGTCGACGACG gcggtctc CGTCGTCAGGATCATCCGGGC")
    Bxb1_attP = lower_no_whitespace("GTCGTGGTTTGTCTGGTCAACCACC gcggtctc AGTGGTGTACGGTACAAACCCCGAC")


class Bxb1Reaction(Reaction):

    @staticmethod
    def allowed():
        return [
        ]


class PhiC31Reaction(Reaction):

    @staticmethod
    def allowed():
        return [
        ]
